@@ -1,3 +1,4 @@
+import asyncio
 import datetime
 import importlib
 import logging
@@ -20,6 +21,7 @@ from slixmpp.plugins.xep_0356.permissions import (
     PresencePermission,
     RosterAccess,
 )
+from sqlalchemy import delete
 
 from slidge import *
 from slidge import (
@@ -33,7 +35,8 @@ from slidge import (
 from slidge.command.categories import ADMINISTRATION
 from slidge.core import config
 from slidge.core.mixins.attachment import AttachmentMixin
-from slidge.util.sql import db
+from slidge.core.session import _sessions
+from slidge.db.models import Contact
 from slidge.util.test import SlidgeTest
 from slidge.util.types import LegacyAttachment, LegacyContactType, LegacyMessageType
 
@@ -44,6 +47,12 @@ unregistered = []
 reactions_received_by_juliet = []
 
 
+class ClearSessionMixin:
+    def tearDown(self):
+        super().tearDown()
+        _sessions.clear()
+
+
 class Gateway(BaseGateway):
     COMPONENT_NAME = "SLIDGE TEST"
 
@@ -52,6 +61,8 @@ class Gateway(BaseGateway):
 
     GROUPS = True
 
+    LEGACY_CONTACT_ID_TYPE = int
+
     async def unregister(self, user: GatewayUser):
         unregistered.append(user)
 
@@ -93,8 +104,8 @@ class Session(BaseSession):
     ):
         if chat.jid_username == "juliet":
             text_received_by_juliet.append((text, chat))
-        assert self.user.bare_jid == "romeo@montague.lit"
-        assert self.user.jid == JID("romeo@montague.lit")
+        assert self.user_jid.bare == "romeo@montague.lit"
+        assert self.user_jid == JID("romeo@montague.lit")
         chat.send_text("I love you")
         return 0
 
@@ -152,14 +163,16 @@ class Bookmarks(LegacyBookmarks):
         await self.by_legacy_id("room2")
 
 
-class Base(SlidgeTest):
+class Base(ClearSessionMixin, SlidgeTest):
     plugin = globals()
 
     def setUp(self):
         super().setUp()
-        user_store.add(
+        user = self.xmpp.store.users.new(
             JID("romeo@montague.lit/gajim"), {"username": "romeo", "city": ""}
         )
+        user.preferences = {"sync_avatar": True, "sync_presence": True}
+        self.xmpp.store.users.update(user)
         self.get_romeo_session().logged = True
 
     @staticmethod
@@ -614,7 +627,7 @@ class TestAimShakespeareBase(Base):
               <query xmlns='http://jabber.org/protocol/disco#items'
                      node='http://jabber.org/protocol/commands'>
                 <item jid="aim.shakespeare.lit"
-                      node="🛷️ Slidge administration"
+                      node="https://slidge.im/command/core/admin"
                       name="🛷️ Slidge administration" />
                 <item jid="aim.shakespeare.lit"
                       node="jabber:iq:register"
@@ -644,16 +657,19 @@ class TestAimShakespeareBase(Base):
               <query xmlns='http://jabber.org/protocol/disco#items'
                      node='http://jabber.org/protocol/commands'>
                 <item jid="aim.shakespeare.lit"
-                      node="🛷️ Slidge administration"
-                      name="🛷️ Slidge administration" />
+                      name="🛷️ Slidge administration"
+                      node="https://slidge.im/command/core/admin" />
                 <item jid="aim.shakespeare.lit"
-                      node="👤 Contacts"
+                      node="https://slidge.im/command/core/contacts"
                       name="👤 Contacts" />
                 <item jid="aim.shakespeare.lit"
-                      node="👥 Groups"
+                      node="https://slidge.im/command/core/groups"
                       name="👥 Groups" />
                 <item jid="aim.shakespeare.lit"
-                      node="unregister"
+                      node="https://slidge.im/command/core/preferences"
+                      name="⚙️ Preferences" />
+                <item jid="aim.shakespeare.lit"
+                      node="https://slidge.im/command/core/unregister"
                       name="❌ Unregister from the gateway" />
               </query>
             </iq>
@@ -681,13 +697,16 @@ class TestAimShakespeareBase(Base):
               <query xmlns='http://jabber.org/protocol/disco#items'
                      node='http://jabber.org/protocol/commands'>
                 <item jid="aim.shakespeare.lit"
-                      node="🛷️ Slidge administration"
+                      node="https://slidge.im/command/core/admin"
                       name="🛷️ Slidge administration" />
                 <item jid="aim.shakespeare.lit"
-                      node="re-login"
+                      node="https://slidge.im/command/core/re-login"
                       name="🔐 Re-login to the legacy network" />
                 <item jid="aim.shakespeare.lit"
-                      node="unregister"
+                      node="https://slidge.im/command/core/preferences"
+                      name="⚙️ Preferences" />
+                <item jid="aim.shakespeare.lit"
+                      node="https://slidge.im/command/core/unregister"
                       name="❌ Unregister from the gateway" />
               </query>
             </iq>
@@ -717,8 +736,8 @@ class TestAimShakespeareBase(Base):
               <query xmlns='http://jabber.org/protocol/disco#items'
                      node='http://jabber.org/protocol/commands'>
                 <item jid="aim.shakespeare.lit"
-                      node="{ADMINISTRATION}"
-                      name="{ADMINISTRATION}" />
+                      node="{ADMINISTRATION.node}"
+                      name="{ADMINISTRATION.name}" />
                 <item jid="aim.shakespeare.lit"
                       node="jabber:iq:register"
                       name="📝 Register to the gateway" />
@@ -729,29 +748,48 @@ class TestAimShakespeareBase(Base):
         config.ADMINS = ()
 
     def test_adhoc_forbidden_non_admin(self):
-        self.recv(  # language=XML
-            f"""
+        with unittest.mock.patch(
+            "slixmpp.plugins.xep_0050.adhoc.XEP_0050.new_session",
+            return_value="session-id",
+        ):
+            self.recv(  # language=XML
+                f"""
             <iq type="set"
                 from="test@localhost/gajim"
                 to="{self.xmpp.boundjid.bare}"
                 id="123">
               <command xmlns="http://jabber.org/protocol/commands"
                        action="execute"
-                       node="{ADMINISTRATION}" />
+                       node="{ADMINISTRATION.node}" />
             </iq>
             """
-        )
+            )
         self.send(  # language=XML
             """
-            <iq xmlns="jabber:component:accept"
-                type="error"
+            <iq type="result"
                 from="aim.shakespeare.lit"
                 to="test@localhost/gajim"
                 id="123">
-              <error type="auth">
-                <subscription-required xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
-                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">User not found</text>
-              </error>
+              <command xmlns="http://jabber.org/protocol/commands"
+                       node="https://slidge.im/command/core/admin"
+                       sessionid="session-id"
+                       status="executing">
+                <actions>
+                  <next />
+                </actions>
+                <x xmlns="jabber:x:data"
+                   type="form">
+                  <title>🛷️ Slidge administration</title>
+                  <field var="command"
+                         label="Command"
+                         type="list-single">
+                    <option label="ℹ️ Server information">
+                      <value>https://slidge.im/command/core/admin/info</value>
+                    </option>
+                    <value />
+                  </field>
+                </x>
+              </command>
             </iq>
             """,
             use_values=False,
@@ -805,6 +843,7 @@ class TestAimShakespeareBase(Base):
                 <feature var="urn:xmpp:hats:0" />
                 <feature var="urn:xmpp:mds:displayed:0+notify" />
                 <feature var="urn:xmpp:mds:displayed:0" />
+                <feature var='urn:xmpp:chat-markers:0' />
               </query>
             </iq>
             """
@@ -941,7 +980,7 @@ class TestAimShakespeareBase(Base):
         juliet: LegacyContact = self.run_coro(
             session.contacts.by_jid(JID("juliet@aim.shakespeare.lit"))
         )
-        juliet.REACTIONS_SINGLE_EMOJI = True
+        LegacyContact.REACTIONS_SINGLE_EMOJI = True
         self.recv(  # language=XML
             f"""
             <iq type="get"
@@ -987,7 +1026,7 @@ class TestAimShakespeareBase(Base):
             </iq>
             """,
         )
-        juliet.REACTIONS_SINGLE_EMOJI = False
+        LegacyContact.REACTIONS_SINGLE_EMOJI = False
 
     def test_non_existing_contact(self):
         self.recv(  # language=XML
@@ -1020,14 +1059,14 @@ class TestAimShakespeareBase(Base):
 
         ids = []
 
-        async def send_file(file_path, legacy_msg_id=None, *_a, **_k):
+        async def send_file(self, file_path, legacy_msg_id=None, *_a, **_k):
             ids.append(legacy_msg_id)
             m = Message()
             m.set_id(legacy_msg_id)
             return "", [m]
 
-        self.juliet.send_file = send_file
-
+        orig = LegacyContact.send_file
+        LegacyContact.send_file = send_file
         self.loop(self.juliet.send_files([], body="Hey"))
         assert not self.next_sent().get_id()
 
@@ -1090,6 +1129,8 @@ class TestAimShakespeareBase(Base):
         self.loop(self.juliet.send_files([a], legacy_msg_id="leg"))
         assert ids.pop(-1) == "leg"
 
+        LegacyContact.send_file = orig
+
     def test_gateway_message(self):
         session = self.get_romeo_session()
         session.send_gateway_message("Hello!")
@@ -1108,12 +1149,12 @@ class TestAimShakespeareBase(Base):
         )
 
 
-class TestPrivilegeOld(SlidgeTest):
+class TestPrivilegeOld(ClearSessionMixin, SlidgeTest):
     plugin = globals()
 
     def setUp(self):
         super().setUp()
-        user_store.add(
+        self.xmpp.store.users.new(
             JID("romeo@shakespeare.lit/gajim"), {"username": "romeo", "city": ""}
         )
 
@@ -1199,12 +1240,12 @@ class TestPrivilegeOld(SlidgeTest):
         )
 
 
-class TestPrivilege(SlidgeTest):
+class TestPrivilege(ClearSessionMixin, SlidgeTest):
     plugin = globals()
 
     def setUp(self):
         super().setUp()
-        user_store.add(
+        self.xmpp.store.users.new(
             JID("romeo@shakespeare.lit/gajim"), {"username": "romeo", "city": ""}
         )
 
@@ -1290,17 +1331,19 @@ class TestPrivilege(SlidgeTest):
         )
 
 
-class TestContact(SlidgeTest):
+class TestContact(ClearSessionMixin, SlidgeTest):
     plugin = globals()
 
     def setUp(self):
         super().setUp()
-        user_store.add(
+        self.xmpp.store.users.new(
             JID("romeo@montague.lit/gajim"), {"username": "romeo", "city": ""}
         )
         self.get_romeo_session().logged = True
         self.get_romeo_session().contacts.ready.set_result(True)
-        db.presence_nuke()
+        with self.xmpp.store.session() as session:
+            session.execute(delete(Contact))
+        # db.presence_nuke()
 
     @staticmethod
     def get_romeo_session() -> Session:
@@ -1345,10 +1388,9 @@ class TestContact(SlidgeTest):
         )
 
     def test_caps_extended(self):
+        LegacyContact.REACTIONS_SINGLE_EMOJI = True
+        LegacyContact.CORRECTION = False
         juliet = self.get_juliet()
-        juliet.REACTIONS_SINGLE_EMOJI = True
-        juliet.CORRECTION = False
-        juliet.reset_caps_cache()
         juliet.is_friend = True
         juliet.online()
         self.send(  # language=XML
@@ -1364,6 +1406,8 @@ class TestContact(SlidgeTest):
             </presence>
             """
         )
+        LegacyContact.REACTIONS_SINGLE_EMOJI = False
+        LegacyContact.CORRECTION = True
 
     def test_vcard_temp(self):
         juliet = self.get_juliet()
@@ -1460,6 +1504,7 @@ class TestContact(SlidgeTest):
     def test_user_subscribe_to_non_friend_accept(self):
         juliet = self.get_juliet()
         juliet.is_friend = False
+
         sub = self.get_presence("subscribe")
 
         with unittest.mock.patch(
@@ -1719,12 +1764,12 @@ class TestContact(SlidgeTest):
         )
 
 
-class TestCarbon(SlidgeTest):
+class TestCarbon(ClearSessionMixin, SlidgeTest):
     plugin = globals()
 
     def setUp(self):
         super().setUp()
-        user_store.add(
+        self.xmpp.store.users.new(
             JID("romeo@shakespeare.lit/gajim"), {"username": "romeo", "city": ""}
         )
         self.recv(  # language=XML
@@ -1902,6 +1947,7 @@ class TestUserGetsOnline(Base):
         super().setUp()
         self.get_romeo_session().contacts.ready.set_result(True)
         self.juliet.is_friend = True
+        self.juliet.added_to_roster = True
         self.juliet.name = "Juliet"
         self.send(  # language=XML
             """
